/* Corr_matrix -- correlation matrix implementation
 *
 * SEE corr_matrix.h for details
 */


/* -- Correlation Matrix includes -- */

#include <stdlib.h>
#ifdef	UNIX
	#include <ieeefp.h>
	#define	FP_MAX	FP_PINF
#endif
#ifdef	WIN32
	#include	<float.h>
	#define	FP_SNAN	_FPCLASS_SNAN
	#define	FP_MAX	DBL_MAX
#endif
#include <memory.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include "corr_matrix.h"
#include "2darray.h"
#include "vector_utils.h"
#include "vector_read.h"
#include "cluster.h"
#include "gauss.h"


/* -- Correlation Matrix defines -- */

/* --- SQR definded in nrutil.h --- */
#define	SQR(x)	((x) * (x))
/* -------------------------------- */


/* -- Correlation Matrix help function definitions -- */

/* -- Function AddCorrelation
 * Pre: 'matrix' is a valid allocated matrix
 *      'correlation' is the partial correlation matrix for a single vector
 * Post: the components of 'correlation' were added to 'matrix'
 */
void AddCorrelation(corrMatrix matrix, double **correlation);

/* -- Function SubtractCorrelation
 * Pre: 'matrix' is a valid allocated matrix
 *      'correlation is the partial correlation matrix for a single vector
 * Post: the components of 'correlation' were subtracted from 'matrix'
 */
void SubtractCorrelation(corrMatrix matrix, double **correlation);

/* -- Function CopyCorrelation
 * Pre: 'matrixd' is an allocated matrix
 *      'matrixs' is a valid allocated matrix to copy
 * Post: the componets of 'matrixs' were copied into 'matrixd'
 */
void CopyCorrelation(double **matrixd, double **matrixs, int dimensions);

/* -- Function UpdateInverse
 * Pre: 'matrix' is avalid allocated correlation matrix
 * Post: the inverse field of 'matrix' has been updated
 */
void UpdateInverse(corrMatrix matrix);

/* -- Function UpdateDummyInverse
 * Pre: 'matrix' is a valid allocated correlation matrix with less than
 *      'MIN_VECTORS_IN_CLUSTER' vectors
 * Post: the inverse field of 'matrix' was updated with an inverse based on the
 *       centroid of 'matrix', but with 'MIN_VECTORS_IN_CLUSTER' vectors.
 */
void UpdateDummyInverse(corrMatrix matrix);

/* -- Function MakeDummyCluster
 * Pre: 'matrix' is a valid allocated correlation matrix
 * Post: (A new cluster was created containing the dummy points used to make 
 *        a dummy correlation matrix && the cluster was returned) ||
 *       (There was a problem making the cluster && NULL was returned)
 */
cluster *MakeDummyCluster(corrMatrix matrix);

/* -- Function CalculateAverage
 * Pre: 'matrix' is a valid allocated correlation matrix
 * Post: The 'avCorrelation' and 'stddevCorr' fields for 'matrix' are now valid &&
 *       'validAverage' == TRUE
 */
void CalculateAverage(corrMatrix matrix);


/* -- Function WriteHex
 * Pre: 'buffer' is an allocated text buffer that 'data' will be printed into in hex
 *      'buffer' must be at least 2 * 'numBytes' + 1 bytes long
 *      'data' is a pointer to the data region to output
 *      'numBytes' is the number of bytes of 'data' to print
 * Post: 'data' was printed in hex to 'buffer'
 */
void WriteHex(char buffer[], const unsigned char *data, size_t numBytes);

/* -- Function ReadHex
 * Pre: 'buffer' is an allocated text buffer that 'data' will be read from
 *      'data' is a pointer to the data region to output (pre-allocated)
 *      'numBytes' is the size of 'data' in bytes'
 * Post: 'data' was printed in hex to 'buffer'
 */
void ReadHex(const char buffer[], unsigned char *data, size_t numBytes);


/* -- Correlation Matrix functions -- */

corrMatrix CMatrixCreate(int dimensions)
{
	corrMatrix	matrix;
	int			xIndex, yIndex;

	if (!(matrix = (corrMatrix) malloc(sizeof(struct CMatrixTag))))
		return NULL;

	if (!Allocate2DArray((char ***) &(matrix -> MATRIX), dimensions, dimensions, sizeof(double))) {
		free(matrix);
		return NULL;
	}

	matrix -> dimensions = dimensions;
	matrix -> num_vectors = 0;
	matrix -> centroid = NULL;
	matrix -> INVERSE = NULL;
	matrix -> validAverage = FALSE;
	matrix -> singular = FALSE;
	matrix -> vector_set = ClusterNew();
	matrix -> isDummy = FALSE;

	for (yIndex = 0; yIndex < dimensions; yIndex++)
		for (xIndex = 0; xIndex < dimensions; xIndex++)
			matrix -> MATRIX[xIndex][yIndex] = 0.0;

	return matrix;
}


corrMatrix CMatrixDestroy(corrMatrix matrix)
{
	CMatrixWipe(matrix);
	Deallocate2DArray((char ***) &(matrix -> MATRIX), matrix -> dimensions);
	if (matrix -> INVERSE)
		Deallocate2DArray((char ***) &(matrix -> INVERSE), matrix -> dimensions + 1);
	free(matrix);
	return NULL;
}


void CMatrixWipe(corrMatrix matrix)
{
	int	xIndex, yIndex;

	if (matrix -> num_vectors == 0)
		return;

	if (matrix -> centroid) matrix -> centroid = VectorDeallocate(matrix -> centroid);
	matrix -> vector_set = ClusterDestroy(matrix -> vector_set);
	matrix -> num_vectors = 0;

	for (yIndex = 0; yIndex < matrix -> dimensions; yIndex++)
		for (xIndex = 0; xIndex < matrix -> dimensions; xIndex++)
			matrix -> MATRIX[xIndex][yIndex] = 0.0;

	if (matrix -> INVERSE) {
		Deallocate2DArray((char ***) &(matrix -> INVERSE), matrix -> dimensions + 1);
		matrix -> INVERSE = NULL;
	}
	matrix -> singular = FALSE;
	matrix -> validAverage = FALSE;
	matrix -> isDummy = FALSE;
}


corrMatrix CMatrixCreateFromCluster(cluster *clust)
{
	corrMatrix	matrix;
	cluster		*cIndex;

	if (!ClusterCheckDimensionality(clust))
		return NULL;

	if (!(matrix = CMatrixCreate(clust -> vect -> dimensions)))
		return NULL;

	cIndex = clust;
	while (cIndex != NULL) {
		if (!CMatrixAddVector(matrix, cIndex -> vect)) {
			matrix = CMatrixDestroy(matrix);
			return NULL;
		}
		cIndex = cIndex -> next;
	}

	return matrix;
}


corrMatrix CMatrixMakeCopy(corrMatrix matrix)
{
	corrMatrix	copy;

	copy = CMatrixCreate(matrix -> dimensions);

	copy -> num_vectors = matrix -> num_vectors;
	copy -> centroid = VectorMakeCopy(matrix -> centroid);
	copy -> vector_set = ClusterMakeCopy(matrix -> vector_set);
	copy -> singular = matrix -> singular;
	CopyCorrelation(copy -> MATRIX, matrix -> MATRIX, copy -> dimensions);
	CopyCorrelation(copy -> INVERSE, matrix -> INVERSE, copy -> dimensions + 1);
	if (matrix -> validAverage) {
		copy -> avCorrelation = matrix -> avCorrelation;
		copy -> stddevCorr = matrix -> stddevCorr;
		copy -> validAverage = TRUE;
	}
	copy -> isDummy = matrix -> isDummy;

	return copy;
}


bool CMatrixAddVector(corrMatrix matrix, vector *vect)
{
	double	**correlation;
	int	dimensions;

	vector	*vectNorm;

	dimensions = vect -> dimensions;
	if (dimensions != matrix -> dimensions)
		return FALSE;

	vectNorm = VectorMakeCopy(vect);

	if (!(correlation = VectorCorrelation(vect, &dimensions)))
		return FALSE;

	AddCorrelation(matrix, correlation);
	matrix -> vector_set = ClusterAddVector(matrix -> vector_set, VectorMakeCopy(vect));
	matrix -> validAverage = FALSE;
	if (matrix -> centroid) {
		VectorDeallocate(matrix -> centroid);
		matrix -> centroid = NULL; /*ClusterCentroid(matrix -> vector_set);*/
	}
	
	Deallocate2DArray((char ***) &correlation, dimensions);
	if (matrix -> INVERSE) {
		Deallocate2DArray((char ***) &(matrix -> INVERSE), matrix -> dimensions + 1);
		matrix -> INVERSE = NULL;
		matrix -> singular = FALSE;
	}

	matrix -> num_vectors++;

	return TRUE;
}

bool CMatrixDeleteVector(corrMatrix matrix, vector *vect)
{
	double	**correlation;
	int	dimensions;

	if (!(correlation = VectorCorrelation(vect, &dimensions)))
		return FALSE;

	if (!ClusterVectorIsIn(matrix -> vector_set, vect))
		return FALSE;

	SubtractCorrelation(matrix, correlation);
	matrix -> vector_set = ClusterRemoveVector(matrix -> vector_set, vect);
	if (matrix -> centroid) {
		VectorDeallocate(matrix -> centroid);
		matrix -> centroid = NULL; /*ClusterCentroid(matrix -> vector_set);*/
	}
	matrix -> validAverage = FALSE;

	Deallocate2DArray((char ***) &correlation, dimensions);
	if (matrix -> INVERSE) {
		Deallocate2DArray((char ***) &(matrix -> INVERSE), matrix -> dimensions + 1);
		matrix -> INVERSE = NULL;
		matrix -> singular = FALSE;
	}

	matrix -> num_vectors--;
	return TRUE;
}


bool CMatrixVectorIsIn(corrMatrix matrix, vector *vect)
{
	return ClusterVectorIsIn(matrix -> vector_set, vect);
}


bool CMatrixIsEmpty(corrMatrix matrix)
{
	return matrix -> num_vectors == 0;
}

void printmatrix(double **matrix, int dimensions)
{
	int	iCol, iRow;

	if (matrix) {
		for (iRow = 0; iRow < dimensions; iRow++) {
			for (iCol = 0; iCol < dimensions; iCol++)
				printf("%12.8f  ", matrix[iRow][iCol]);
			printf("\n");
		}
	} else
		printf("{NULL}\n");
}

double CMatrixCorrelationWithVect(corrMatrix matrix, vector *vect)
{
	double	sum;
	int	dimensions,
			iCol, iRow;

	vector	*vectH;

	if (CMatrixIsEmpty(matrix))
		return FP_MAX;

	if (!matrix -> INVERSE)
		UpdateInverse(matrix);

	if (matrix -> singular)					/* Singular matrix; we can't calculate */
		return FP_SNAN;                  /* the correlation, so return NAN      */

	dimensions = matrix -> dimensions;

	vectH = VectorMakeHomogeneous(vect);

	sum = 0.0;
	for (iRow = 0; iRow < dimensions + 1; iRow++) {
		for (iCol = 0; iCol < dimensions + 1; iCol++)
			sum += (vectH -> components[iRow]) * (matrix -> INVERSE[iCol][iRow]) * (vectH -> components[iCol]);
	}

	VectorDeallocate(vectH);

	return sum;
}


double CMatrixNormedCorrWithVect(corrMatrix matrix, vector *vect)
{
	if (CMatrixIsEmpty(matrix))
		return FP_MAX;

/*	if (matrix -> num_vectors < 3) {
		if (!matrix -> centroid)
			matrix -> centroid = ClusterCentroid(matrix -> vector_set);

		return VectorEuclideanDist(vect, matrix -> centroid);
	}
*/
	if (!matrix -> validAverage) {
		CalculateAverage(matrix);
		matrix -> validAverage = TRUE;
	}
	
	return CMatrixCorrelationWithVect(matrix, vect) / matrix -> avCorrelation;
}

double CMatrixNormWithVector(corrMatrix matrix, vector *vect)
{
	double	*product,
				sum,
				num_vectors;
	int	dimensions,
			iCol, iRow;

	if (CMatrixIsEmpty(matrix))
		return FP_MAX;

	dimensions = matrix -> dimensions;
	num_vectors = (double) matrix -> num_vectors;
/*
	printf("Norming matrix:\n");
	printmatrix(matrix -> MATRIX, dimensions);
	printf("--------------\n");
	printmatrix(matrix -> INVERSE, dimensions);
	printf("With vector: ");
	VectorWrite(stdout, vect);
	printf("\n");
*/
	if (!matrix -> INVERSE)
		UpdateInverse(matrix);

	if (matrix -> singular)					/* Singular matrix, we can't calculate */
		return FP_SNAN;                  /* the 2-norm, so return NAN           */

	if (!(product = (double *) malloc((dimensions + 1) * sizeof(double))))
		return FP_SNAN;

	for (iRow = 0; iRow < dimensions + 1; iRow++) {
		sum = 0.0;
		for (iCol = 0; iCol < dimensions; iCol++)
			sum += (matrix -> INVERSE[iCol][iRow] * vect -> components[iCol]);
		sum += matrix -> INVERSE[dimensions][iRow]; /* * 1.0; */
		product[iRow] = sum;
	}

	sum = 0.0;
	for (iRow = 0; iRow < dimensions + 1; iRow++)
		sum += SQR(product[iRow]);

	sum = sqrt(sum);

	free(product);
	return sum;
}

/*
	product[dimensions] = 0.0;
	for (yIndex = 0; yIndex < dimensions; yIndex++) {
		sum = 0.0;
		for (xIndex = 0; xIndex < dimensions; xIndex++)
			sum += (matrix -> MATRIX[xIndex][yIndex] / (double) matrix -> num_vectors) * vect -> components[xIndex];
		sum += matrix -> centroid -> components[yIndex];
		product[yIndex] = sum;
		product[dimensions] += matrix -> centroid -> components[yIndex] /* * (double) matrix -> num_vectors * vect -> components[yIndex];
	}

	product[dimensions] += 1.0;

	sum = 0.0;
	for (yIndex = 0; yIndex <= dimensions; yIndex++)
		sum += SQR(product[yIndex]); /* / product[dimensions]);

	sum += 1.0;
	sum = sqrt(sum);

	free(product);
	return sum;
}*/


double **CMatrixGetConceptualMatrix(corrMatrix matrix, int *matrixSize)
{
	double	**concepMatrix;
	int		iCol, iRow;

	*matrixSize = matrix -> dimensions + 1;

	if (!Allocate2DArray((char ***) &concepMatrix, *matrixSize, *matrixSize, sizeof(double)))
		return NULL;
	
	if (!matrix -> centroid)
		matrix -> centroid = ClusterCentroid(matrix -> vector_set);
	
	for (iRow = 0; iRow < *matrixSize - 1; iRow++) {
		for (iCol = 0; iCol < *matrixSize - 1; iCol++)
			concepMatrix[iCol][iRow] = matrix -> MATRIX[iCol][iRow] / (double) matrix -> num_vectors;
		concepMatrix[*matrixSize - 1][iRow] = matrix -> centroid -> components[iRow];
	}

	for (iCol = 0; iCol < *matrixSize - 1; iCol++)
		concepMatrix[iCol][*matrixSize - 1] = matrix -> centroid -> components[iCol];

	concepMatrix[*matrixSize - 1][*matrixSize - 1] = 1.0;

	return concepMatrix;
}


void CMatrixWrite(FILE *output, corrMatrix matrix)
{
	int	iRow, iCol,
			dimensions;
	char	buffer[50];
	vector	*centroid;

	dimensions = matrix -> dimensions;

	fprintf(output, "d: %d\n", dimensions);
	fprintf(output, "nv: %d\n", matrix -> num_vectors);
	
	if (!matrix -> validAverage)
		CalculateAverage(matrix);

	WriteHex(buffer, (char *) &(matrix -> avCorrelation), sizeof(double));
	fprintf(output, "ac: %s\n", buffer); 
	WriteHex(buffer, (char *) &(matrix -> stddevCorr), sizeof(double));
	fprintf(output, "sdc: %s\n", buffer);
	fprintf(output, "sing: %d\n", matrix -> singular);
	
	centroid = CMatrixCentroid(matrix);
	
	fprintf(output, "c: ");

	for (iRow = 0; iRow < dimensions; iRow++) {
		WriteHex(buffer, (char *) &(centroid -> components[iRow]), sizeof(double));
		fprintf(output, "%s ", buffer);
	}
	fprintf(output, "\nMATRIX\n");

	for (iRow = 0; iRow < dimensions; iRow++) {
		for (iCol = 0; iCol < dimensions; iCol++) {
			WriteHex(buffer, (char *) &(matrix -> MATRIX[iRow][iCol]), sizeof(double));
			fprintf(output, "%s ", buffer);
		}
		fprintf(output, "\n");
	}
}

corrMatrix CMatrixRead(FILE *input)
{
	corrMatrix	newMatrix;
	int			iRow, iCol, dimensions;
	char			buffer[50];
	vector		*centroid;

	if (fscanf(input, " d: %d", &dimensions) == 0)
		return NULL;

	newMatrix = CMatrixCreate(dimensions);

	fscanf(input, " nv: %d", &(newMatrix -> num_vectors));
	newMatrix -> validAverage = TRUE;
	fscanf(input, " ac: %s", buffer);
	ReadHex(buffer, (char *) &(newMatrix -> avCorrelation), sizeof(double));
	fscanf(input, " sdc: %s", buffer);
	ReadHex(buffer, (char *) &(newMatrix -> stddevCorr), sizeof(double));

	fscanf(input, " sing: %d", &(newMatrix -> singular));

	if (!(centroid = VectorAllocate(dimensions))) {
		CMatrixDestroy(newMatrix);
		return NULL;
	}

	fscanf(input, " c: ");
	for (iRow = 0; iRow < dimensions; iRow++) {
		fscanf(input, "%s ", buffer);
		ReadHex(buffer, (char *) &(centroid -> components[iRow]), sizeof(double));
	}

	newMatrix -> centroid = centroid;

	if (!Allocate2DArray((char ***) &(newMatrix -> MATRIX), dimensions, dimensions, sizeof(double))) {
		CMatrixDestroy(newMatrix);
		return NULL;
	}

	fscanf(input, " MATRIX ");
	for (iRow = 0; iRow < dimensions; iRow++) {
		for (iCol = 0; iCol < dimensions; iCol++) {
			fscanf(input, "%s ", buffer);
			ReadHex(buffer, (char *) &(newMatrix -> MATRIX[iRow][iCol]), sizeof(double));
		}
	}

	return newMatrix;
}

void CMatrixWriteMatlab(FILE *output, corrMatrix matrix)
{
	int	dimensions,
			iCol, iRow;
	double	**concepMatrix;

	if (!(concepMatrix = CMatrixGetConceptualMatrix(matrix, &dimensions))) {
		fprintf(stderr, "*** Error getting conceptual matrix.\n");
		return;
	}

	fprintf(output, "[");
	for (iRow = 0; iRow < dimensions; iRow++) {
		for (iCol = 0; iCol < dimensions; iCol++) {
			fprintf(output, "%f", concepMatrix[iCol][iRow]);
			if (iCol < dimensions - 1)
				fprintf(output, ",");
		}
		if (iRow < dimensions - 1)
			fprintf(output, ";");
	}
	fprintf(output, "]");
} 

double CMatrixRMSCorrelation(corrMatrix matrix)
{
        double  corrSum;
        cluster *cIndex;
                
        if (CMatrixIsEmpty(matrix))
                return 0.0;
                 
        corrSum = 0.0;
        cIndex = matrix -> vector_set;
        while (cIndex != NULL) {
                corrSum += SQR(CMatrixCorrelationWithVect(matrix, cIndex -> vect));
                cIndex = cIndex -> next;
        }
        
/*        corrSum /= (double) matrix -> num_vectors;*/
        return sqrt(corrSum);
}

double CMatrixAvCorrWithinCluster(corrMatrix matrix)
{
	if (!matrix -> validAverage)
		CalculateAverage(matrix);

	return matrix -> avCorrelation;
}


double CMatrixStdDevCorrWithinCluster(corrMatrix matrix)
{
	if (!matrix -> validAverage)
		CalculateAverage(matrix);

	return matrix -> stddevCorr;
}

vector *CMatrixCentroid(corrMatrix matrix)
{
	if (!matrix -> centroid)
		matrix -> centroid = ClusterCentroid(matrix -> vector_set);

	return matrix -> centroid;
}

bool CMatrixMerge(corrMatrix dest, corrMatrix source)
{
	cluster	*cIndex;

	if (source -> num_vectors == 0)
		return TRUE;

	cIndex = source -> vector_set;
	while (cIndex != NULL) {
		if (!CMatrixAddVector(dest, VectorMakeCopy(cIndex -> vect)))
			return FALSE;

		cIndex = cIndex -> next;
	}

	return TRUE;
}


bool CMatrixUseDummy(corrMatrix matrix)
{
	if (matrix -> num_vectors < MIN_VECTORS_IN_CLUSTER)
		return TRUE;
	else
		return FALSE;
}


/* -- Correlation Matrix helper functions -- */
void AddCorrelation(corrMatrix matrix, double **correlation)
{
	int	dimensions,
		xIndex, yIndex;

	dimensions = matrix -> dimensions;

	for (yIndex = 0; yIndex < dimensions; yIndex++)
		for (xIndex = 0; xIndex < dimensions; xIndex++)
			matrix -> MATRIX[xIndex][yIndex] += correlation[xIndex][yIndex];
}


void SubtractCorrelation(corrMatrix matrix, double **correlation)
{
	int	dimensions,
		xIndex, yIndex;

	dimensions = matrix -> dimensions;

	for (yIndex = 0; yIndex < dimensions; yIndex++)
		for (xIndex = 0; xIndex < dimensions; xIndex++)
			matrix -> MATRIX[xIndex][yIndex] -= correlation[xIndex][yIndex];
}

void CopyCorrelation(double **matrixd, double **matrixs, int dimensions)
{
	int	xIndex, yIndex;

	for (yIndex = 0; yIndex < dimensions; yIndex++)
		for (xIndex = 0; xIndex < dimensions; xIndex++)
			matrixd[xIndex][yIndex] = matrixs[xIndex][yIndex];
}

void UpdateInverse(corrMatrix matrix)
{
	double	**conceptual;
	int		dimINVERSE, dimConceptual;

	dimINVERSE  = matrix -> dimensions + 1;

	if (matrix -> INVERSE)
		Deallocate2DArray((char ***) &(matrix -> INVERSE), dimINVERSE);

	if (!CMatrixIsEmpty(matrix)) {
		if (!(matrix -> isDummy) && CMatrixUseDummy(matrix)) {		/* Less than x vectors, do dummy */
			UpdateDummyInverse(matrix);

		} else {
			if (!(conceptual = CMatrixGetConceptualMatrix(matrix, &dimConceptual)))
				return;

			if (!Allocate2DArray((char ***) &(matrix -> INVERSE), dimINVERSE, dimINVERSE, sizeof(double))) {
				fprintf(stderr, "--- 2DArray allocation failed for matrix inverse.\n");
				return;
			}

			CopyCorrelation(matrix -> INVERSE, conceptual, dimINVERSE);
			if (!GaussJordanInverse(conceptual, matrix -> INVERSE, dimINVERSE)) {
				/*fprintf(stderr, "--- Warning: Matrix was singular.\n");*/
				matrix -> singular = TRUE;
			}
			Deallocate2DArray((char ***) &conceptual, dimConceptual);
		}
	}
}

void UpdateDummyInverse(corrMatrix matrix)
{
	corrMatrix	dummyMatrix;
	int		numDimensions;
	cluster	*dummy, *clustIndex;

	numDimensions = matrix -> dimensions;
	dummyMatrix = CMatrixCreate(numDimensions);

	if (!matrix -> centroid)
		matrix -> centroid = ClusterCentroid(matrix -> vector_set);

	clustIndex = dummy = MakeDummyCluster(matrix);
	while (clustIndex != NULL) {
		CMatrixAddVector(dummyMatrix, clustIndex -> vect);
		clustIndex = clustIndex -> next;
	}

	dummyMatrix -> isDummy = TRUE;
	UpdateInverse(dummyMatrix);

	if (!(matrix -> INVERSE)) {
		Allocate2DArray(	(char ***) &(matrix -> INVERSE),
								numDimensions + 1, numDimensions + 1,
								sizeof(double));
	}

	CopyCorrelation(matrix -> INVERSE, dummyMatrix -> INVERSE, numDimensions + 1);

	CMatrixDestroy(dummyMatrix);
	ClusterDestroy(dummy);
}

cluster *MakeDummyCluster(corrMatrix matrix)
{
	int		numDimensions, dimIndex;
	double	av_distance,
				*centroid,
				*dummy;
	cluster	*dummyCluster, *clustIndex;
	vector	*dummyVect;

	numDimensions = matrix -> dimensions;

	if (!matrix -> centroid)
		matrix -> centroid = ClusterCentroid(matrix -> vector_set);

	if (matrix -> num_vectors == 1) {
		av_distance = INITIAL_DISTANCE;
	} else {
		clustIndex = matrix -> vector_set;
		av_distance = 0.0;
		while (clustIndex != NULL) {
			av_distance += VectorEuclideanDist(clustIndex -> vect, matrix -> centroid);
			clustIndex = clustIndex -> next;
		}
		av_distance /= (double) matrix -> num_vectors;
	}	

	dummyCluster = ClusterNew();
	centroid = matrix -> centroid -> components;
	dummy = (double *) malloc (numDimensions * sizeof(double));
	dimIndex = 0;
	while (dimIndex < numDimensions) {
		memcpy(dummy, centroid, numDimensions * sizeof(double));
		dummy[dimIndex] += av_distance;										/* +ve along dimension */
		dummyVect = VectorConstruct(numDimensions, dummy);
		dummyCluster = ClusterAddVector(dummyCluster, VectorMakeCopy(dummyVect));
		VectorDeallocate(dummyVect);

		memcpy(dummy, centroid, numDimensions * sizeof(double));
		dummy[dimIndex] -= av_distance;										/* -ve along dimension */
		dummyVect = VectorConstruct(numDimensions, dummy);
		dummyCluster = ClusterAddVector(dummyCluster, VectorMakeCopy(dummyVect));
		VectorDeallocate(dummyVect);

		dimIndex++;
	}

	free(dummy);
	return dummyCluster;
}


void CalculateAverage(corrMatrix matrix)
{
	double	correlation, sum, sumSqr;
	int		numVects;
	cluster	*dummyCluster, *cIndex;

	if (CMatrixUseDummy(matrix)) {
		cIndex = dummyCluster = MakeDummyCluster(matrix);
	} else {
		cIndex = matrix -> vector_set;
	}

	sum = 0.0;
	sumSqr = 0.0;
	numVects = 0;
	while (cIndex != NULL) {
		correlation =  CMatrixCorrelationWithVect(matrix, cIndex -> vect);
		sum += correlation;
		sumSqr += SQR(correlation);
		cIndex = cIndex -> next;
		numVects++;
	}

	if (CMatrixUseDummy(matrix)) {
		ClusterDestroy(dummyCluster);
	}

	matrix -> avCorrelation = sum / (double) numVects;
	matrix -> stddevCorr = ((double) numVects * sumSqr) - SQR(sum);
	matrix -> stddevCorr /= (double) SQR(numVects);
	matrix -> stddevCorr = sqrt(matrix -> stddevCorr);

	matrix -> validAverage = TRUE;
}

void WriteHex(char buffer[], const unsigned char *data, size_t numBytes)
{
	size_t	byteIndex;
	char		hex[5];

	buffer[0] = 0x0;			/* Buffer initially null */
	for (byteIndex = 0; byteIndex < numBytes; byteIndex++) {
		if (data[byteIndex] < 0x10)
			sprintf(hex, "0%x", data[byteIndex]);
		else
			sprintf(hex, "%x", data[byteIndex]);	/* Get the hex character */
		strcat(buffer, hex);
	}
}

void ReadHex(const char buffer[], unsigned char *data, size_t numBytes)
{
	size_t		byteIndex;
	char			hex[5];
	short int	dec;

	for (byteIndex = 0; byteIndex < numBytes; byteIndex++) {
		hex[0] = '0';
		hex[1] = 'x';
		hex[2] = buffer[byteIndex * 2];
		hex[3] = buffer[byteIndex * 2 + 1];
		hex[4] = 0x0;

		sscanf(hex, "%x", &dec);
		data[byteIndex] = (unsigned char) dec;
	}
}


/* --- END of corr_matrix.c --- */
