/* Corr_matrix -- correlation maxtrix implementation for 2nd order clustering
 *
 * Author: Dylan Muir (dr.muir@student.qut.edu.au)
 *         (c) QUT MLRC Semester 2 1999
 *         (c) QUT MLRC Semester 1 2000
 *         (c) QUT MLRC Semester 2 2000
 * Date: 5th November, 1999
 * Modified: 15th December, 2000 (Post-GST)
 * Version: 0.21
 *
 * Aknowledgements: Lipson, H., Siegelmann, H. T., "High Order Shape Neurons for Data Structure Decomposition"
 *
 * Uses: Vector_utils, Cluster, Stddefs
 */

#ifndef	__corr_matrix
#define	__corr_matrix

/* -- Correlation matrix includes -- */

#include <stdlib.h>
#include "cluster.h"
#include "vector_utils.h"
#include "stddefs.h"

#ifdef	WIN32
	#include	<float.h>
	#define	FP_SNAN	_FPCLASS_SNAN
#endif

#define	MIN_VECTORS_IN_CLUSTER	8		/* approximate cluster with a dummy until we have x vectors */
#define	INITIAL_DISTANCE			0.1	/* for an approximate cluster, how far away  */
													/* from the centroid to put the dummy points */

typedef struct CMatrixTag {
	int	dimensions,			/* Dimensionality of vectors */
			num_vectors;		/* Used for normalising MATRIX */
	double	**MATRIX,		/* Covariance matrix [dim x dim] */
				**INVERSE,		/* Correlation matrix inverse [dim + 1 x dim + 1] */
				avCorrelation,	/* Average correlation from all vectors within cluster */
				stddevCorr;		/* Standard deviation correlation for all vectors within cluster */
	bool		singular,		/* Matrix is singular */
				validAverage,	/* Average correlation is valid (otherwise calculate it) */
				isDummy;			/* Is this matrix a dummy cluster? */
	vector	*centroid;		/* Euclidean centroid of cluster */
	cluster	*vector_set;	/* Set of vectors comprising cluster */
} *corrMatrix;

/* Notes:
 *
 * The full (conceptual) correlation matrix contains vectors in homogeneous
 * form, and augments the correlation terms with what becomes effectively the
 * cluster centroids along the right side and on the bottom.  The full matrix
 * is also in normalised form.
 *
 * Rh = Sum (cluster) [X Trans(X)]   <-- Trans == transpose (i.e. vector product)
 *
 *    = Sum [a][a b c 1]
 *          [b]
 *          [c]
 *          [1]  <-- Homogeneous form adds this to every pattern vector
 *
 *    = Sum (j) [a2 ab ac | a] = Rh(j)
 *              [ba b2 bc | b]
 *              [ca cb c2 | c]
 *              [---------+  ]
 *              [ a  b  c   1]
 *
 * This matrix is then normalised by dividing by size(cluster)
 * The correlation of a vector with (cluster) is given by finding the
 * product of X * the inverse of Rh * Trans(X)
 */

/* -- Correlation matrix functions -- */

/* -- Function CMatrixCreate
 * Pre: 'dimensions' > 0
 * Post: (A new empty correlation matrix was created and returned) ||
 *       (There was a problem with allocation && NULL was returned)
 */
corrMatrix CMatrixCreate(int dimensions);

/* -- Function CMatrixDestroy
 * Pre: 'matrix' is a valid allocated correlation matrix
 * Post: 'matrix' was destroyed && a null matrix was returned
 */
corrMatrix CMatrixDestroy(corrMatrix matrix);

/* -- Function CMatrixWipe
 * Pre: 'matrix' is a valid allocated matrix
 * Post: 'matrix' was cleared so that it contains no vectors
 */
void CMatrixWipe(corrMatrix matrix);

/* -- Function CMatrixCreateFromCluster
 * Pre: 'clust' is a valid cluster
 * Post: (a new matrix was created from 'clust' and returned) ||
 *       (there was a problem with allocation && NULL was returned)
 */
corrMatrix CMatrixCreateFromCluster(cluster *clust);

/* -- Function CMatrixCreateFromVectorSet
 * Pre: 'vectorSet' is an array of allocated vectors
 *      'numVectors' is the number of vectors in 'vectorSet'
 * Post: (a new matrix was created from COPIES of the vectors in 'vectorSet' and returned) ||
 *       (there was a problem with allocation && NULL was returned)
 */
corrMatrix CMatrixCreateFromVectorSet(vector *vectorSet[], int numVectors);

/* -- Function CMatrixMakeCopy
 * Pre: 'matrix' is a valid allocated matrix
 * Post: (a new matrix was copied from 'matrix and returned) ||
 *       (there was a problem with allocation && NULL was reurned)
 */
corrMatrix CMatrixMakeCopy(corrMatrix matrix);

/* -- Function CMatrixAddVector
 * Pre: 'matrix' is a valid allocated matrix
 *      'vect' is a valid allocated vector
 * Post: ('vect' was added to 'matrix' && TRUE was returned) ||
 *       ('vect' could not be added to 'matrix' && FALSE was returned)
 * Note: This functions adds a COPY of 'vect'
 */
bool CMatrixAddVector(corrMatrix matrix, vector *vect);

/* -- Function CMatrixDeleteVector
 * Pre: 'matrix' is a valid allocated matrix
 *      'vect' is a valid allocated vector
 * Post: ('vect' was in 'matrix' && 'vect' was removed from 'matrix' && TRUE was returned) ||
 *       ('vect' was not in 'matrix' && FALSE was returned)
 */
bool CMatrixDeleteVector(corrMatrix matrix, vector *vect);

/* -- Function CMatrixVectorIsIn
 * Pre: 'matrix' is a valid allocated matrix
 *      'vect' is a valid allocated vector
 * Post: ('vect' exists in 'matrix' && TRUE was returned) ||
 *       ('vect' does not exist in 'matrix' && FALSE was returned)
 */
bool CMatrixVectorIsIn(corrMatrix matrix, vector *vect);

/* -- Function CMatrixIsEmpty
 * Pre: 'matirx' is a valid allocated matrix
 * Post: ('matrix' contains no vectors & TRUE was returned) ||
 *       ('matrix' contains at least one vector && FALSE was returned)
 */
bool CMatrixIsEmpty(corrMatrix matrix);

/* -- Function CMatrixCorrelationWithVect
 * Pre: 'matrix' is a valid non-empty allocated matrix
 *      'vect' is a valid allocated vector
 * Post: If 'matrix' contains at least three vectors, it will be treated as
 *       a cluster and the correlation will be calculated.  If the matrix contains
 *       less than three vectors, the Euclidean distance to the centroid (a spherical
 *       measure) will be returned instead.  If 'matrix' is singular, FP_SNAN will be
 *       returned.
 */
double CMatrixCorrelationWithVect(corrMatrix matrix, vector *vect);

/* -- Function CMatrixNormedCorrWithVect
 * Pre: 'matrix' is a valid non-empty allocated matrix
 *      'vect' is a valid allocated vector
 * Post: Returns the normalised correlation from 'vect' -> 'matrix'
 */
double CMatrixNormedCorrWithVect(corrMatrix matrix, vector *vect);

/* -- Function CMatrixNormWithVector
 * Pre: 'matrix' is a valid non-empty allocated matrix
 *      'vect' is a valid allocated vector
 * Post: The euclidean norm of the vector product of 'vect' and 'matrix' was
 *       returned (See top of file for description).  If 'matrix' is singular, FP_SNAN
 *       will be returned.
 */
double CMatrixNormWithVector(corrMatrix matrix, vector *vect);

/* -- Function CMatrixGetConceptualMatrix
 * Pre: 'matrix' is a valid allocated matrix
 *      'matrixSize' points to an allocated integer
 * Post: The conceptual (real) correlation matrix is returned as a 2D
 *       double array.  The matrix is square; the size is placed in 'matrixSize'
 *       NULL is returned if allocation was not possible.  In this case,
 *       'matrixSize' is undefined
 */
double **CMatrixGetConceptualMatrix(corrMatrix matrix, int *matrixSize);

/* -- Function CMatrixWrite
 * Pre: 'output' is a file opened for writing
 *      'matrix' is a valid allocated matrix
 * Post: the matrix of 'matrix' was written to output in a way that can be read with
 *        the CMatrixRead function
 */
void CMatrixWrite(FILE *output, corrMatrix matrix);

/* -- Function CMatrixRead
 * Pre: 'input' is a file opened for writing
 * Post: (a new matrix was read from 'input' and returned) &&
 *       (there was a problem reading the file && NULL was returned)
 */
corrMatrix CMatrixRead(FILE *input);

/* -- Function CMatrixWriteConceptual
 * Pre: 'output' is a file opened for writing
 *      'matrix' is a valid allocated matrix
 * Post: the conceptual matrix of 'matrix' was written to output
 */
void CMatrixWriteConceptual(FILE *output, corrMatrix matrix);

/* -- Function CMatrixWriteMatlab
 * Pre: 'output' is a file opened for writing
 *      'matrix' is a valid allocated matrix
 * Post: the conceptual matrix of 'matrix' was written to output in a format
 *       that is easy to copy into matlab
 */
void CMatrixWriteMatlab(FILE *output, corrMatrix matrix);

/* -- Function CMatrixRMSCorrelation
 * Pre: 'matrix' is a valid allocated matrix
 * Post: the RMS correlation over all vectors in 'matrix' was computed and returned
 */
double CMatrixRMSCorrelation(corrMatrix matrix);

/* -- Fnuction CMatrixAvCorrWithinCluster
 * Pre: 'matrix' is a valid allocated matrix with >= 3 vectors
 * Post: The average correlation from all vectors within cluster to the matrix was
 *       calculated and returned
 */
double CMatrixAvCorrWithinCluster(corrMatrix matrix);

/* -- Fnuction CMatrixStdDevCorrWithinCluster
 * Pre: 'matrix' is a valid allocated matrix with >= 3 vectors
 * Post: The standard deviation from all vectors within cluster to the matrix was
 *       calculated and returned
 */
double CMatrixStdDevCorrWithinCluster(corrMatrix matrix);

/* -- Function CMatrixCentroid
 * Pre: 'matrix' is a valid allocated matrix
 * Post: 'matrix's centroid was returned as an EXISTING vector: must be duplicated
 */
vector *CMatrixCentroid(corrMatrix matrix);

/* -- Function CMatrixMerge
 * Pre: 'dest' is a valid corrMatrix
 *      'source' is a valid corrMatrix
 * Post: (All the vectors in 'source' were COPIED into 'dest' && TRUE was returned) ||
 *       (there was a problem with allocation && FALSE was returned)
 */
bool CMatrixMerge(corrMatrix dest, corrMatrix source);

/* -- Function CMatrixUseDummy
 * Pre: 'matrix' is a valid matrix
 * Post: ('matrix' doesn't meet the criteria for a cluster && returned TRUE) ||
 *       (returned FALSE)
 */
bool CMatrixUseDummy(corrMatrix matrix);


#endif	__corr_matrix

/* --- END of corr_matrix.h --- */
