/* smdarray -- Sparse, multi-dimensional arrays
 *
 * see smdarray.h for details
 */

/* -- smdarray includes -- */

#include "smdarray.h"
#include "stddefs.h"
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>


/* -- smdarray helper functions prototypes -- */

/* -- Function sArrayEnsureCellExistsDiscrete
 * Pre: 'array' is a valid created smd array
 *      'numDimensions' is the number of dimensions to access in 'array'
 *      'indices' is an integer array of the indices, of order 'numDimensions'
 * Post: (a cell representing 'array[indices]' was returned) ||
 *       (there was a problem with allocation && NULL was returned)
 */
smdarray sArrayEnsureCellExistsDiscrete(smdarray array, int numDimensions, int indices[]);

/* -- Function sArrayDoesCellExistDiscrete
 * Pre: 'array' is a valid created smd array
 *      'numDimensions' is the number of dimensions to access in 'array'
 *      'indices' is an integer array of the indices, of order 'numDimensions'
 * Post: ('array[indices]' has had a alue assigned to it && TRUE was returned) ||
 *       ('array[indices]' has not had a value assigned to it && FALSE was returned)
 */
bool sArrayDoesCellExistDiscrete(smdarray array, int numDimensions, int indices[]);

/* -- Function sArrayQueryCellDiscrete
 * Pre: 'array' is a valid created smd array
 *      'numDimensions' is the number of dimensions to access in 'array'
 *      'indices' is an integer array of the indices, of order 'numDimensions'
 * Post: (a cell representing 'array[indices]' was returned if one exists) ||
 *       (the cell doesn't exists && NULL was returned)
 */
smdarray sArrayQueryCellDiscrete(smdarray array, int numDimensions, int indices[]);


/* -- Function sArrayMakeCell
 * Pre: 'index' and 'value' are reasonable integers
 * Post: (a cell was created, 'index' and 'value' were assigned to it, the
 *       nextCell and nextDimension indices were set to NULL && the cell was returned) ||
 *       (there was a problem with allocation && NULL was returned)
 */
smdarray sArrayMakeCell(int index, int value);


/* -- smdarray functions -- */

smdarray sArrayCreate(void)
{
	smdarray	array;

	return sArrayMakeCell(0, 0);
}

int sArrayAccess(smdarray array, int numDimensions, ...)
{
	va_list	pindex;
	int	*indices,
		dIndex;

	if (!(indices = (int *) malloc(sizeof(int) * numDimensions)))
		return 0;

	va_start(pindex, numDimensions);
	dIndex = 0;
	while (dIndex < numDimensions) {
		indices[dIndex] = va_arg(pindex, int);
		dIndex++;
	}
	va_end(pindex);

	if (!sArrayDoesCellExistDiscrete(array, numDimensions, indices))
		return 0;

	return sArrayQueryCellDiscrete(array, numDimensions, indices) -> cell;
}

smdarray sArrayGetCell(smdarray array, int numDimensions, ...)
{
	va_list	pindex;
	int	*indices,
		dIndex;

	if (!(indices = (int *) malloc(sizeof(int) * numDimensions)))
		return NULL;

        va_start(pindex, numDimensions);
        dIndex = 0;
        while (dIndex < numDimensions) {
                indices[dIndex] = va_arg(pindex, int);
                dIndex++;
        }
        va_end(pindex);

	return sArrayEnsureCellExistsDiscrete(array, numDimensions, indices);
}

bool sArrayDoesCellExist(smdarray array, int numDimensions, ...)
{
        va_list pindex;
        int     *indices,
                dIndex;

        if (!(indices = (int *) malloc(sizeof(int) * numDimensions)))
                return FALSE;

        va_start(pindex, numDimensions);
        dIndex = 0;
        while (dIndex < numDimensions) {
                indices[dIndex] = va_arg(pindex, int);
                dIndex++;
        }
        va_end(pindex);

	return sArrayDoesCellExistDiscrete(array, numDimensions, indices);
}

bool sArrayAssign(smdarray array, int numDimensions, int value, ...)
{
        va_list 	pindex;
        int     	*indices,
                	dIndex;
	smdarray	cell;

        if (!(indices = (int *) malloc(sizeof(int) * numDimensions)))
                return FALSE;

        va_start(pindex, value);
        dIndex = 0;
        while (dIndex < numDimensions) {
                indices[dIndex] = va_arg(pindex, int);
                dIndex++;
        }
        va_end(pindex);

	if (!(cell = sArrayEnsureCellExistsDiscrete(array, numDimensions, indices)))
		return FALSE;

	cell -> cell = value;
	return TRUE;
}

smdarray sArrayDestroy(smdarray array)
{
	if (array -> nextDimension != NULL)
		array -> nextDimension = sArrayDestroy(array -> nextDimension);
	if (array -> nextCell != NULL)
		array -> nextCell = sArrayDestroy(array -> nextCell);

	free(array);
	return NULL;
}

bool sArrayIncrement(smdarray array, int numDimensions, int delta, ...)
{
        va_list 	pindex;
        int     	*indices,
                	dIndex;
	smdarray	cell;

        if (!(indices = (int *) malloc(sizeof(int) * numDimensions)))
                return FALSE;

        va_start(pindex, delta);
        dIndex = 0;
        while (dIndex < numDimensions) {
                indices[dIndex] = va_arg(pindex, int);
                dIndex++;
        }
        va_end(pindex);

	if (!(cell = sArrayEnsureCellExistsDiscrete(array, numDimensions, indices)))
		return FALSE;

	cell -> cell += delta;
	return TRUE;
}

int sArrayCountLeaves(smdarray cell)
{
	int	count;

	count = 0;

	if (cell -> nextDimension == NULL) {
		/* This cell is a leaf */
		count += 1;
	} else {
		/* This cell is not a leaf, branch down dimensions */
		count += sArrayCountLeaves(cell -> nextDimension);
	}

	if (cell -> nextCell != NULL) {
		/* Continue to count leaves in this dimension */
		count += sArrayCountLeaves(cell -> nextCell);
	}

	return count;
}

int sArrayBranchWeight(smdarray cell)
{
	int	weight;

	weight = 0;

	if (cell -> nextDimension == NULL) {
		/* This cell is a leaf, so add to the weight */
		weight += cell -> cell;
	} else {
		/* This cell is not a leaf, branch down dimensions */
		weight += sArrayBranchWeight(cell -> nextDimension);
	}

	if (cell -> nextCell != NULL) {
		/* Continue to count weight in this dimension */
		weight += sArrayBranchWeight(cell -> nextCell);
	}

	return weight;
}

/* -- smdarray helper functions -- */

bool sArrayDoesCellExistDiscrete(smdarray array, int numDimensions, int indices[])
{
	return (sArrayQueryCellDiscrete(array, numDimensions, indices) != NULL);
}

smdarray sArrayQueryCellDiscrete(smdarray array, int numDimensions, int indices[])
{
	int		dIndex;
	smdarray	cell;

	cell = array;
	dIndex = 0;
	while (dIndex < numDimensions) {
		while (cell -> index < indices[dIndex]) {
			if (cell -> nextCell == NULL)
				return NULL;
			cell = cell -> nextCell;
		}
		if (cell -> index != indices[dIndex])
			return NULL;
		dIndex++;
		if (dIndex < numDimensions) {
			if (cell -> nextDimension == NULL)
				return NULL;
			cell = cell -> nextDimension;
		}
	}
	return cell;
}


smdarray sArrayEnsureCellExistsDiscrete(smdarray array, int numDimensions, int indices[])
{
	int		dIndex;
	smdarray	cell, newCell;

	if (sArrayDoesCellExistDiscrete(array, numDimensions, indices))
		return sArrayQueryCellDiscrete(array, numDimensions, indices);

	cell = array;
	dIndex = 0;
	while (dIndex < numDimensions) {
		while (cell -> index < indices[dIndex]) {
			if ((cell -> nextCell == NULL) || (cell -> nextCell -> index > indices[dIndex])) {
				if (!(newCell = sArrayMakeCell(indices[dIndex], 0)))
					return NULL;
				newCell -> nextCell = cell -> nextCell;
				cell -> nextCell = newCell;
			}
			cell = cell -> nextCell;
		}
		dIndex++;
		if (dIndex < numDimensions) {
			if ((cell -> nextDimension == NULL) || (cell -> nextDimension -> index > indices[dIndex])) {
				if (!(newCell = sArrayMakeCell(indices[dIndex], 0)))
					return NULL;
				newCell -> nextCell = cell -> nextDimension;
				cell -> nextDimension = newCell;
			}
			cell = cell -> nextDimension;
		}
	}

	if (!sArrayDoesCellExistDiscrete(array, numDimensions, indices)) {
		printf("houston, we have a problem.\n");
		return NULL;
	}
	return sArrayQueryCellDiscrete(array, numDimensions, indices);
}

smdarray sArrayMakeCell(int index, int value)
{
	smdarray	newCell;

	if (!(newCell = (smdarray) malloc(sizeof(struct smdArrayTag))))
		return NULL;

	newCell -> index = index;
	newCell -> cell = value;
	newCell -> nextCell = NULL;
	newCell -> nextDimension = NULL;

	return newCell;
}


/* --- END of smdarray.c --- */


