/* smdarray -- Sparse Multi-Dimensional Arrays
 *
 * Author: Dylan Muir (dr.muir@student.qut.edu.au)
 *         QUT MLRC LPG Semester 2 1999
 * Date: 22nd September, 1999
 * Modified: 27th September, 1999 (hbtm)
 *           6th October, 1999
 * Version: 0.04
 *
 * Description: This module allows for the use of infinitely-dimensioned
 *              sparse arrays.  You'll have problems in C when you start
 *              doing 4- and 5-gram transition tables of 29-class data.
 *              Allocating the array is HUGE (29 * 29 * 29 * 29 = 707281
 *              cells for 4-gram) and most of the elements won't even be
 *              used!  This module provides the solution.
 */

#ifndef	__smdarray_h
#define	__smdarray_h

/* -- smdarray defines -- */

#define	smd_MAX_DIMENSIONS	100

/* -- smdarray required modules -- */

#include "stddefs.h"


/* -- smdarray structure -- */

typedef struct smdArrayTag {
	int	index, 				/* index in current dimension */
			cell;					/* contents of current cell */

	struct smdArrayTag	*nextCell,			/* iterator within dimension */
								*nextDimension;	/* pointer to higher dimension */
} *smdarray;					/* default typename to use */


/* -- smdArray functions -- */

/* -- Function sArrayCreate
 * Pre: TRUE
 * Post: returns an empty smdarray or NULL if creation failed
 */
smdarray sArrayCreate(void);

/* -- Function sArrayDestroy
 * Pre: 'array' is a valid smdarray
 * Post: all cells in 'array' were destroyed && an empty array was returned
 */
smdarray sArrayDestroy(smdarray array);

/* -- Function sArrayAccess
 * Pre: 'array' is  avalid created smd array
 *      'numDimensions' is the total number of dimensions you are accessing
 *      '...' is a list of indices into 'array'
 * Post: (a value has been assigned to 'array[...]' && the value was returned) ||
 *       (a value has not been assigned to 'array[...]' && zero was returned)
 */
int sArrayAccess(smdarray array, int numDimensions, ...);

/* -- Function sArrayGetCell
 * Pre: 'array' is a valid created smd array
 *      'numDimensions' is the number of dimensions to access in 'array'
 *      '...' is a list of indices into 'array'
 * Post: (a cell corresponding to 'array[...]' was returned) ||
 *       (there was a problem with allocation && NULL was returned)
 * Note: Since sArrayGetCell will create a cell if one doesn't exist, indexing
 *       through 'array' using sArrayGetCell would be exceedingly stupid, since
 *       when you finished you wouldn't have a sparse array anymore.
 *       You probably shouldn't index along .nextCell and .nextDimension.
 */
smdarray sArrayGetCell(smdarray array, int numDimensions, ...);

/* -- Funcion sArrayDoesCellExist
 * Pre: 'array' is a valid created smd array
 *      'numDimensions' is the number of dimensions to access in 'array'
 *      '...' is a list of indices into 'array'
 * Post: ('array[...]' has had a value assigned to it && TRUE was returned) ||
 *       ('array[...]' has never had a value assigned to it && FALSE was returned))
 */
bool sArrayDoesCellExist(smdarray array, int numDimensions, ...);

/* -- Function sArrayAssign
 * Pre: 'array' is a valid created smd array
 *      'numDimensions' is the number of dimensions to access in 'array'
 *      'value' is the value to assign to 'array[...]'
 *      '...' is a list of indices into 'array'
 * Post: ('value' was assigned to 'array[...]' && TRUE was returned) ||
 *       (there was a problem with allocation && FALSE was returned)
 */
bool sArrayAssign(smdarray array, int numDimensions, int value, ...);

/* -- Function sArrayIncrement
 * Pre: 'array' is a valid created smd array
 *      'numDimensions' is the number of dimensions to access in 'array'
 *      'delta' is a number to add arithmetically to 'array[...]'
 *      '...' is a list of indices into 'array'
 * Post: ('array[...]' was incremented by 'delta' && TRUE was returned) ||
 *       (there was a problem with allocation && FALSE was returned)
 */
bool sArrayIncrement(smdarray array, int numDimensions, int delta, ...);

/* -- Function sArrayCountLeaves
 * Pre: 'cell' is a valid cell in an smd array
 * Post: the number of leaves in subdimensions off 'cell' was returned
 */
int sArrayCountLeaves(smdarray cell);

/* -- Function sArrayBranchWeight
 * Pre: 'cell' is a valid cell in an smd array
 * Post: the values of all the leaves of 'cell' were summed and returned
 */
int sArrayBranchWeight(smdarray cell);

#endif	/* __smdarray_h */

/* --- END of smdarray.h --- */


